from dataclasses import dataclass
from typing import Any

import numpy as np
from sklearn.base import ClassifierMixin, RegressorMixin
from skrub import DataOp
from skrub._data_ops._data_ops import Apply, GetItem
from skrub._data_ops._evaluation import find_node, find_X, find_y


@dataclass
class DagSummary:
    task_type: str | None = None
    model: str | None = None
    model_name: str | None = None
    model_steps: str | None = None
    model_definition: str | None = None
    target_name: str | None = None
    target_definition: str | None = None
    target_steps: str | None = None
    target_unique_values_from_preview: list[Any] | None = None
    dataset_description: str | None = None


# TODO We can do much more here!
def _summarise_dag(dag_sink_node: DataOp) -> DagSummary:
    summary = DagSummary()

    def is_model(some_op):
        if hasattr(some_op, "_skrub_impl"):
            impl = some_op._skrub_impl
            if isinstance(impl, Apply) and hasattr(impl, "estimator"):
                est = impl.estimator
                return isinstance(est, (ClassifierMixin, RegressorMixin))
        return False

    model_node = find_node(dag_sink_node, predicate=is_model)
    if model_node is not None:
        estimator = model_node._skrub_impl.estimator
        summary.model_steps = model_node.skb.describe_steps()
        summary.model_definition = model_node._skrub_impl.creation_stack_description()
        summary.model = f"{estimator.__class__.__module__}.{estimator.__class__.__qualname__}"
        if isinstance(estimator, ClassifierMixin):
            summary.task_type = "classification"
        if isinstance(estimator, RegressorMixin):
            summary.task_type = "regression"

    y_op = find_y(dag_sink_node)
    if y_op is not None:
        summary.target_steps = y_op.skb.describe_steps()
        if hasattr(y_op, "_skrub_impl"):
            summary.target_definition = y_op._skrub_impl.creation_stack_description()
            if y_op.skb.name is not None:
                summary.target_name = y_op.skb.name
            elif isinstance(y_op._skrub_impl, GetItem):
                summary.target_name = y_op._skrub_impl.key
            try:
                summary.target_unique_values_from_preview = [val.item() for val in np.unique(y_op.skb.preview())]
            except Exception as __e:  # pylint: disable=broad-exception-caught
                pass

    X_op = find_X(dag_sink_node)
    if X_op is not None:
        if X_op.skb.description is not None:
            summary.dataset_description = X_op.skb.description

    return summary



import builtins
import re
from collections.abc import Iterable
from typing import Any

import numpy
import pandas as pd
import sklearn
import skrub

_ALLOWED_MODULES = ["numpy", "pandas", "sklearn", "skrub", "re"]


def _make_safe_import(allowed_modules: Iterable[str]):
    real_import = builtins.__import__

    def safe_import(name, globals_to_import=None, locals_to_import=None, fromlist=(), level=0):
        top_name = name.split(".")[0]
        if top_name not in allowed_modules:
            raise ImportError(f"Import of '{name}' is not allowed")
        return real_import(name, globals_to_import, locals_to_import, fromlist, level)

    return safe_import


def _safe_exec(
    python_code: str,
    variable_to_return: str,
    safe_locals_to_add: dict[str, Any] | None = None,
) -> Any:
    if safe_locals_to_add is None:
        safe_locals_to_add = {}

    safe_builtins = {
        "__import__": _make_safe_import(_ALLOWED_MODULES),
        "int": int,
        "float": float,
        "str": str,
        "bool": bool,
        "list": list,
        "dict": dict,
        "set": set,
        "tuple": tuple,
        "len": len,
        "range": range,
        "isinstance": isinstance,
        "sum": sum,
        "any": any,
        "all": all,
        "map": map,
        "hash": hash,
    }

    safe_globals = {
        "__builtins__": safe_builtins,
        "skrub": skrub,
        "sklearn": sklearn,
        "numpy": numpy,
        "np": numpy,
        "pandas": pd,
        "pd": pd,
        "re": re,
    }

    safe_locals = safe_locals_to_add
    exec(python_code, safe_globals, safe_locals_to_add)

    return safe_locals[variable_to_return]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from sklearn.base import BaseEstimator, TransformerMixin
from skrub import DataOp
from skrub.selectors._base import Filter

from gyyre.optimisers._dag_summary import DagSummary


class GyyreContextAwareMixin(ABC):
    """
    A mixin for gyyre operators that want to adjust themselves to the context in which they are used.

    The context is captured in a DagSummary object, which is provided to the operator as constructor parameter.
    The DagSummary contains information about the overall task, the model being used, the target variable, and the
    computational graph.

    Attributes:
        gyyre_dag_summary (DagSummary | None): A summary of the computational graph context.
    """

    gyyre_dag_summary: DagSummary | None = None


class GyyrePrefittableMixin(ABC):
    """
    A mixin for gyyre operators that can export and import their internal state for skipping the fit operation.

    The prefitted state is captured in a dict, which is provided to the operator as constructor parameter.
    This is required to enable context-aware optimisation, where the operator is fitted multiple times in
    different contexts, and where the pre-fitted state has to be evaluated repeatedly during cross-validation

    Attributes:
        gyyre_prefitted_state (dict[str, Any] | None): The prefitted state of the operator.
    """

    gyyre_prefitted_state: dict[str, Any] = {}

    @abstractmethod
    def state_after_fit(self) -> dict[str, Any]:
        """
        Return the internal state of the operator after fitting, to be used for prefitting in future fits.

        Returns:
            dict[str, Any]: The internal state of the operator.
        """


class GyyreOptimisableMixin(ABC):
    gyyre_memory: list[dict[str, Any]] = {}

    @abstractmethod
    def memory_update_from_latest_fit(self) -> dict[str, Any]:
        pass


class SemChooseOperator(ABC):
    @abstractmethod
    def set_params_on_estimator(
        self,
        data_op: DataOp,
        estimator: BaseEstimator,
        choices,
        y=None,
    ) -> None:
        """Set parameters on the given estimator based on choices provided."""


class SemSelectOperator(ABC):
    @abstractmethod
    def generate_column_selector(
        self,
        nl_prompt: str,
    ) -> Filter:
        """Generate a column selector for dataframes."""


class WithSemFeaturesOperator(ABC):
    @abstractmethod
    def generate_features_estimator(
        self,
        data_op: DataOp,
        nl_prompt: str,
        name: str,
        how_many: int,
    ) -> BaseEstimator & TransformerMixin & GyyreContextAwareMixin & GyyrePrefittableMixin & GyyreOptimisableMixin:
        """Return an estimator that computes features on a pandas df."""


class SemFillNAOperator(ABC):
    @abstractmethod
    def generate_imputation_estimator(
        self,
        data_op: DataOp,
        target_column: str,
        nl_prompt: str,
    ) -> BaseEstimator & TransformerMixin:
        """Return an estimator that imputes missing values for the target column on a pandas df."""



from typing import Any

from skrub import _dataframe as sbd
from skrub.selectors._base import Filter

from gyyre._code_gen._exec import _safe_exec
from gyyre._code_gen._llm import _generate_python_code
from gyyre._operators import SemSelectOperator


def _sem_select(column: Any, nl_prompt: str) -> bool:
    column_name = sbd.name(column)

    try:
        size = sbd.shape(column)[0]
        sample_size = min(10, size)
        sample = list(sbd.sample(column, sample_size))

        numeric_profile = ""
        if sbd.is_numeric(column):
            numeric_profile = f"""
        ### Numerical information    
        - minimum: {sbd.min(column)}
        - maximum: {sbd.max(column)}
        - mean: {sbd.mean(column)}            
            """

        prompt = f"""

        Your goal is to help a data scientist select relevant columns from a pandas dataframe for a machine learning script 
        which they are developing. You make your decision for each column individually. 

        The data scientist wants you to take special care to the following: 
        {nl_prompt}

        Here is detailed information about a column for which you need to make a decision now:

        ### Details about the column
        - name: {column_name}
        - sample_of_ten_or_less_randomly_chosen_values: {sample}
        - num_unique_values: {sbd.n_unique(column)}
        - has_nulls ? {sbd.has_nulls(column)}
        - is_all_null ? {sbd.is_all_null(column)}
        - is_sorted ? {sbd.is_sorted(column)}

        ### Type information about the column
        - is_bool ? {sbd.is_bool(column)}
        - is_numeric ? {sbd.is_numeric(column)}
        - is_integer ? {sbd.is_integer(column)}
        - is_float ? {sbd.is_float(column)}
        - is_string ? {sbd.is_string(column)}
        - is_object ? {sbd.is_object(column)}
        - is_any_date ? {sbd.is_any_date(column)}
        - is_duration ? {sbd.is_duration(column)}
        - is_categorical ? {sbd.is_categorical(column)}

        {numeric_profile}

        Please respond with a piece of Python code that expresses your decision as a boolean. 
        ONLY RESPOND WITH THE CODE BLOCK, NOTHING ELSE.

        ```python
        __column_is_relevant = <True or False>
        ```            
        """

        python_code = _generate_python_code(prompt)
        decision = _safe_exec(python_code, "__column_is_relevant")

        return decision

    except Exception as e:  # pylint: disable=broad-except
        print(f"An error occurred for column {column_name}:", e)
        return False


class SemSelectLLM(SemSelectOperator):
    def generate_column_selector(
        self,
        nl_prompt: str,
    ) -> Filter:
        return Filter(_sem_select, args=(nl_prompt,), name="sem_select")



from sklearn.base import BaseEstimator
from skrub import selectors
from skrub import DataOp
from skrub._data_ops._skrub_namespace import SkrubNamespace

from gyyre._operator_impls._sem_choose_llm import SemChooseLLM
from gyyre._operator_impls._sem_select_llm import SemSelectLLM
from gyyre._operator_impls._with_sem_features_caafe import WithSemFeaturesCaafe
from gyyre._operator_impls._sem_fillna_llm_plus_model import SemFillNALLLMPlusModel
from gyyre._operator_impls._sem_fillna_llm import SemFillNAWithLLLM
from gyyre.optimisers.greedy import greedy_optimise_semantic_operator


def sem_choose(**kwargs) -> dict:
    return kwargs


def apply_with_sem_choose(
    self: DataOp,
    estimator: BaseEstimator,
    *,
    y=None,
    cols=selectors.all(),
    exclude_cols=None,
    how: str = "auto",
    allow_reject: bool = False,
    unsupervised: bool = False,
    choices=None,
):
    data_op = self
    SemChooseLLM().set_params_on_estimator(data_op, estimator, choices, y=y)
    # TODO forward the * args
    return self.apply(
        estimator,
        y=y,
        cols=cols,
        exclude_cols=exclude_cols,
        how=how,
        allow_reject=allow_reject,
        unsupervised=unsupervised,
    )


def with_sem_features(
    self: DataOp,
    nl_prompt: str,
    name: str,
    how_many: int = 10,
) -> DataOp:
    data_op = self
    feature_gen_estimator = WithSemFeaturesCaafe().generate_features_estimator(data_op, nl_prompt, name, how_many)
    return self.skb.apply(feature_gen_estimator).skb.set_name(name)


def sem_fillna(
    self: DataOp,
    target_column: str,
    nl_prompt: str,
    impute_with_existing_values_only: bool,
    **kwargs,
) -> DataOp:
    data_op = self

    if "with_llm_only" in kwargs and kwargs["with_llm_only"]:
        imputation_estimator = SemFillNAWithLLLM().generate_imputation_estimator(
            data_op, target_column, nl_prompt, impute_with_existing_values_only
        )
    else:
        # TODO Handle this case better for users
        assert impute_with_existing_values_only
        imputation_estimator = SemFillNALLLMPlusModel().generate_imputation_estimator(data_op, target_column, nl_prompt)
    return self.skb.apply(imputation_estimator)


def sem_select(
    self: DataOp,
    nl_prompt: str,
) -> DataOp:
    selector = SemSelectLLM().generate_column_selector(nl_prompt)
    return self.skb.select(selector)


DataOp.with_sem_features = with_sem_features
DataOp.sem_fillna = sem_fillna
DataOp.sem_select = sem_select
SkrubNamespace.apply_with_sem_choose = apply_with_sem_choose

__all__ = ["sem_choose", "greedy_optimise_semantic_operator"]



import traceback
from collections.abc import Iterable

from sklearn.base import BaseEstimator
from skrub import DataOp

from gyyre._code_gen._exec import _safe_exec
from gyyre._code_gen._llm import _generate_python_code
from gyyre._operators import SemChooseOperator


class SemChooseLLM(SemChooseOperator):
    def set_params_on_estimator(
        self, data_op: DataOp, estimator: BaseEstimator, choices: dict[str, str] | None, y=None
    ) -> None:
        print(f"--- gyyre.apply_with_sem_choose({estimator}, {choices})")

        max_retries = 3

        if choices is not None:
            for param_name, user_prompt in choices.items():
                previous_exceptions = []
                for attempt in range(1, max_retries + 1):
                    try:
                        prompt = self._build_prompt(
                            estimator, user_prompt, param_name, previous_exceptions=previous_exceptions
                        )
                        python_code = _generate_python_code(prompt)

                        suggested_choices = _safe_exec(python_code, "__generated_sempipes_choices")
                        estimator.set_params(**{param_name: suggested_choices})
                        print(f"\tSuggested choices for {param_name}: {suggested_choices}")
                        break
                    except Exception as e:  # pylint: disable=broad-except
                        print(f"An error occurred in attempt {attempt}:", e)
                        tb_str = traceback.format_exc()
                        previous_exceptions.append(tb_str)

    @staticmethod
    def _build_prompt(
        estimator: BaseEstimator,
        user_prompt: str,
        param_name: str,
        previous_exceptions: Iterable[Exception] | None = None,
    ) -> str:
        previous_exceptions_memory = ""
        if previous_exceptions is None:
            previous_exceptions = []

        if len(previous_exceptions) > 0:
            previous_exceptions_memory = """
            # PREVIOUS FAILURES:
            This request has been previously attempted, but the generated code did not work. 
            Here are the previous exceptions:\n
            """
            previous_exceptions_memory += "\n".join([f"### Previous exception: {str(e)}" for e in previous_exceptions])

        estimator_name = f"{estimator.__class__.__module__}.{estimator.__class__.__name__}"

        return f"""
        You need to help a data scientist improve their machine learning script in Python, which is written 
        using Pandas, sklearn and skrub. They are using scikit-learn==1.7.1, pandas==2.3.2 and skrub==0.6.1 
    
        # HERE IS AN EXAMPLE:
    
        The data scientist needs help with choosing the `penalty` of a `sklearn.linear_model.LogisticRegression`. 
        Their specific request is:
    
        ```Regularizers that work well for high-dimensional data.```  
    
        This is an example of a valid response:
    
        __generated_sempipes_choices = skrub.choose_from(['l1', 'elasticnet'])
    
        # HERE IS ANOTHER EXAMPLE:
    
        The data scientist needs help with choosing the `high_cardinality` of a `skrub.TableVectorizer`. 
        Their specific request is:
    
        ```Non-embedding based techniques to encode text data.```  
    
        This is an example of a valid response:
    
        __generated_sempipes_choices = skrub.choose_from([
            skrub.MinHashEncoder(n_components=30, ngram_range=(2, 4)),
            skrub.GapEncoder(n_components=10, batch_size=1024)
        ])
    
        {previous_exceptions_memory}
    
        # HERE COMES THE ACTUAL REQUEST:
    
        The data scientist needs help with choosing the `{param_name}` of a `{estimator_name}`. 
        Their specific request is:
    
        ```{user_prompt}```
    
        Please respond with Python code that specifies parameter values that match their request. Make sure that 
        the Python code is executable and uses the correct types. IMPORTANT: The first suggested choice should be 
        the most likely to work well, and is very critical to get right since it will often be used as a default value! 
    
        ONLY INCLUDE VALID PYTHON CODE IN YOUR RESPONSE, NO MARKDOWN OR TEXT. 
        
        THE PYTHON CODE SHOULD MATCH THIS TEMPLATE:
    
        __generated_sempipes_choices = skrub.choose_from([<Your suggested parameter values>])
        """



from litellm import batch_completion, completion

_DEFAULT_MODEL = "openai/gpt-4.1"


def _unwrap_python(text: str) -> str:
    prefix = "```python"
    suffix = "```"
    suffix2 = "```end"
    text = text.strip()
    if text.startswith(prefix):
        text = text[len(prefix) :]
    if text.endswith(suffix):
        text = text[: -len(suffix)]
    if text.endswith(suffix2):
        text = text[: -len(suffix2)]
    text = text.strip()
    # remove lines starting with ``` or ```python (ignoring leading spaces)
    lines = text.splitlines(keepends=True)
    keep = [ln for ln in lines if not ln.lstrip().startswith("```")]
    return "".join(keep)


def _get_cleaning_message(prompt: str) -> list[dict]:
    return [
        {
            "role": "system",
            "content": "You are a helpful assistant, assisting data scientists with data cleaning and preparation, for instance, completing and imputing their data.",
        },
        {"role": "user", "content": prompt},
    ]


def _generate_python_code(prompt: str) -> str:
    messages = [
        {
            "role": "system",
            "content": "You are a helpful assistant, assisting data scientists with completing and improving their machine learning and data preparation code.",
        },
        {"role": "user", "content": prompt},
    ]
    return _generate_python_code_from_messages(messages)


def _generate_python_code_from_messages(messages: list[dict]) -> str:
    print(f"\t> Querying '{_DEFAULT_MODEL}' with {len(messages)} messages...'")

    response = completion(
        model=_DEFAULT_MODEL,
        messages=messages,
        temperature=0.0,
    )

    # TODO add proper error handling
    raw_code = response.choices[0].message["content"]
    return _unwrap_python(raw_code)


def _batch_generate_results(
    prompts: list[str],
    batch_size: int,
) -> list[str | None]:
    """
    Calls litellm.batch_completion with one message-list per prompt.
    Returns a list of raw strings aligned with `prompts`.
    """
    assert batch_size is not None and batch_size > 0, "batch_size must be a positive integer"

    all_messages = [_get_cleaning_message(prompt) for prompt in prompts]
    outputs = []

    for start_index in range(0, len(prompts), batch_size):
        message_batch = all_messages[start_index : start_index + batch_size]

        responses = batch_completion(
            model=_DEFAULT_MODEL,
            messages=message_batch,
            temperature=0.0,
        )

        for response in responses:
            try:
                content = response.choices[0].message["content"]
                outputs.append(content)
            except Exception as e:
                print(f"Error processing response: {response}")
                raise e

    return outputs



from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from gyyre._code_gen._llm import _batch_generate_results
from gyyre._operators import SemFillNAOperator


class SemFillNAWithLLLM(SemFillNAOperator):
    def generate_imputation_estimator(
        self,
        data_op,
        target_column,
        nl_prompt,
        impute_with_existing_values_only=True,
    ):
        return LLMImputer(
            target_column=target_column,
            nl_prompt=nl_prompt,
            impute_with_existing_values_only=impute_with_existing_values_only,
        )


class LLMImputer(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        target_column: str,
        nl_prompt: str,
        impute_with_existing_values_only: bool,
    ) -> None:
        self.target_column = target_column
        self.nl_prompt = nl_prompt
        self.impute_with_existing_values_only = impute_with_existing_values_only
        self.target_column_type_ = None
        self.target_column_unique_values_ = None

    @staticmethod
    def _build_prompt(
        target_column,
        target_column_type,
        candidate_columns,
        nl_prompt,
        target_column_unique_values,
        impute_existing_values_only,
    ):
        # TODO handle this intelligently
        assert len(target_column_unique_values) < 100, "Too many unique values to include in the prompt."
        target_column_unique_values = ", ".join([str(value) for value in target_column_unique_values])

        if impute_existing_values_only:
            values_instruction = f""" The target column has the following unique values: {target_column_unique_values}. Use only existing values to fill in the missing values. """
        else:
            values_instruction = f" Example values from the {target_column} column: {target_column_unique_values}. "

        return f"""
        The data scientist wants to fill missing values in the column '{target_column}' of type '{target_column_type}' in a dataframe. 
        The dataframe has the following columns available with these values to help with this task: {candidate_columns}. You need to assist the data scientist with filling the missing values in the target column.
        {values_instruction}
        
        The data scientist wants you to take special care to the following: 
        {nl_prompt}.
        
        Answer with only the value to impute, without any additional text or formatting.
        Result:    
        """

    def fit(self, df, y=None):  # pylint: disable=unused-argument
        print(f"--- Sempipes.sem_fillna_llm('{self.target_column}', '{self.nl_prompt}')")

        self.target_column_type_ = df[self.target_column].dtype
        self.target_column_unique_values_ = list(df[self.target_column].dropna().unique())
        # TODO maybe add examples of good imputations

        return self

    def transform(self, df):
        check_is_fitted(self, ["target_column_type_", "target_column_unique_values_"])

        indices_to_impute = df[self.target_column].isna()
        rows_to_impute = df[indices_to_impute]

        if rows_to_impute.shape[0] == 0:
            print("\t> No missing values to impute.")
            return df

        # TODO Should we check if there are more previously unseen unique values in the target column?

        # Build prompts for each row to impute
        prompts = []
        for __, row in rows_to_impute.iterrows():
            candidate_columns = {column: row[column] for column in df.columns if column != self.target_column}
            prompt = self._build_prompt(
                self.target_column,
                self.target_column_type_,
                candidate_columns,
                self.nl_prompt,
                self.target_column_unique_values_,
                self.impute_with_existing_values_only,
            )
            prompts.append(prompt)

        # TODO find a way to set the batch size
        results = _batch_generate_results(prompts, batch_size=10)

        df.loc[indices_to_impute, self.target_column] = results

        print(f"\t> Imputed {len(results)} values...")

        return df



from collections.abc import Iterable
from typing import Self

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.utils.validation import check_is_fitted
from skrub import DataOp

from gyyre._code_gen._exec import _safe_exec
from gyyre._code_gen._llm import _generate_python_code
from gyyre._operators import SemFillNAOperator


class SemFillNALLLMPlusModel(SemFillNAOperator):
    def generate_imputation_estimator(self, data_op: DataOp, target_column: str, nl_prompt: str):
        # TODO explore computational graph or cached preview results to improve imputer generation
        return LearnedImputer(target_column, nl_prompt)


class LearnedImputer(BaseEstimator, TransformerMixin):
    def __init__(self, target_column: str, nl_prompt: str):
        self.target_column = target_column
        self.nl_prompt = nl_prompt
        self.imputation_model_ = None

    @staticmethod
    def _build_prompt(
        target_column: str, target_column_type: str, candidate_columns: Iterable[str], nl_prompt: str
    ) -> str:
        return f"""
        The data scientist wants to fill missing values in the column '{target_column}' of type '{target_column_type}' 
        in a dataframe. The dataframe has the following columns available to help with this task: 
        {candidate_columns}. 
        
        You need to assist the data scientists with choosing which columns to use to fill the missing values in 
        the target column. The data scientist wants you to take special care to the following: 
        {nl_prompt}.

        Code formatting for your answer:
        ```python
        __chosen_columns = [<subset of `candidate_columns`>]
        ```end

        The codeblock ends with ```end and starts with "```python"
    Codeblock:    
    """

    def fit(self, df: pd.DataFrame, y=None) -> Self:
        print(f"--- gyyre.sem_fillna('{self.target_column}', '{self.nl_prompt}')")

        target_column_type = str(df[self.target_column].dtype)
        candidate_columns = [column for column in df.columns if column != self.target_column]

        prompt = self._build_prompt(self.target_column, target_column_type, candidate_columns, self.nl_prompt)
        python_code = _generate_python_code(prompt)
        feature_columns = _safe_exec(python_code, "__chosen_columns")

        X = df[feature_columns]
        y = df[self.target_column]

        num_cols = X.select_dtypes(include=[np.number]).columns.tolist()
        cat_cols = X.select_dtypes(exclude=[np.number]).columns.tolist()

        preprocess = ColumnTransformer(
            transformers=[
                (
                    "num",
                    Pipeline([("impute", SimpleImputer(strategy="median")), ("scale", StandardScaler())]),
                    num_cols,
                ),
                (
                    "cat",
                    Pipeline(
                        [
                            ("impute", SimpleImputer(strategy="most_frequent")),
                            ("oh", OneHotEncoder(handle_unknown="ignore")),
                        ]
                    ),
                    cat_cols,
                ),
            ],
            remainder="drop",
        )

        is_numeric_target = pd.api.types.is_numeric_dtype(y)
        if is_numeric_target:
            learner = RandomForestRegressor(random_state=0)
        else:
            learner = RandomForestClassifier(random_state=0)

        model = Pipeline([("prep", preprocess), ("est", learner)])

        # TODO we could keep a small holdout set to measure the imputer performance
        # Train on rows where target is known
        known_mask = y.notna()
        print(f"\t> Fitting imputation model {learner} on columns {feature_columns} of {known_mask.sum()} rows...")
        model.fit(X[known_mask], y[known_mask])
        self.imputation_model_ = model

        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        check_is_fitted(self, "imputation_model_")
        y = df[self.target_column]
        missing_mask = y.isna()
        num_missing_values = missing_mask.sum()

        if num_missing_values > 0:
            print(f"\t> Imputing {num_missing_values} values...")
            df.loc[missing_mask, self.target_column] = self.imputation_model_.predict(df[missing_mask])
        return df



# TODO This class needs some serious cleanup / refactoring
from typing import Any

import pandas as pd
import skrub
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted
from skrub import DataOp

from gyyre._code_gen._exec import _safe_exec
from gyyre._code_gen._llm import _generate_python_code_from_messages
from gyyre._operators import (
    GyyreContextAwareMixin,
    GyyreOptimisableMixin,
    GyyrePrefittableMixin,
    WithSemFeaturesOperator,
)
from gyyre.optimisers._dag_summary import DagSummary

_MAX_RETRIES = 5
_SYSTEM_PROMPT = (
    "You are an expert datascientist assistant solving Kaggle problems. "
    "You answer only by generating code. Answer as concisely as possible."
)


def _get_prompt(
    df: pd.DataFrame,
    nl_prompt: str,
    how_many: int,
    samples: str | None = None,
    dag_summary: DagSummary | None = None,
) -> str:
    data_description_unparsed = None

    task_description = (
        "This code generates additional columns that are useful for a downstream classification "
        "algorithm (such as XGBoost) predicting a target label."
    )
    usefulness = ""
    model_reference = "classifier"

    if dag_summary is not None:
        task_type = dag_summary.task_type
        model = dag_summary.model
        target_name = dag_summary.target_name
        data_description_unparsed = dag_summary.dataset_description

        if task_type and model and target_name:
            task_description = (
                f"This code generates additional columns that are useful for a "
                f'downstream {task_type} algorithm ({model}) predicting "{target_name}".'
            )

        if task_type and target_name:
            action = "predict"
            if task_type == "classification":
                action = "classify"
            usefulness = (
                f"\n# Usefulness: (Description why this adds useful real world knowledge "
                f'to {action} "{target_name}" according to dataset description and attributes.)'
            )

        if task_type == "regression":
            model_reference = "regressor"

    return f"""
The dataframe `df` is loaded and in memory. Columns are also named attributes.
Description of the dataset in `df` (column dtypes might be inaccurate):
"{data_description_unparsed}"

Columns in `df` (true feature dtypes listed here, categoricals encoded as int):
{samples}

This code was written by an expert datascientist working to improve predictions. It is a snippet of code that adds 
new columns to the dataset. Number of samples (rows) in training dataset: {int(len(df))}

{task_description}
Additional columns add new semantic information, that is they use real world knowledge on the dataset. They can e.g. 
be feature combinations, transformations, aggregations where the new column is a function of the existing columns.
The scale of columns and offset does not matter. Make sure all used columns exist. Follow the above description of 
columns closely and consider the datatypes and meanings of classes.
The {model_reference} will be trained on the dataset with the generated columns and evaluated on a holdout set. The 
evaluation metric is accuracy. The best performing code will be selected.
Added columns can be used in other codeblocks.

The data scientist wants you to take special care to the following: {nl_prompt}.


Code formatting for each added column:
```python
# (Feature name and description){usefulness}
# Input samples: (Three samples of the columns used in the following code, e.g. '{df.columns[0]}':
{list(df.iloc[:3, 0].values)}, '{df.columns[1]}': {list(df.iloc[:3, 1].values)}, ...)
(Some pandas code using {df.columns[0]}', '{df.columns[1]}', ... to add a new column for each row in df)
```end

Each codeblock generates up to {how_many} useful columns. Generate as many features as useful for downstream 
{model_reference}, but as few as necessary to reach good performance.
Each codeblock ends with ```end and starts with "```python"
Codeblock:
"""


def _build_prompt_from_df(
    df: pd.DataFrame, nl_prompt: str, how_many: int, dag_summary: DagSummary | None = None
) -> str:
    samples = ""
    df_ = df.head(10)
    for column in list(df_):
        null_ratio = df[column].isna().mean()
        nan_freq = f"{null_ratio * 100:.2g}"
        sampled_values = df_[column].tolist()
        if str(df[column].dtype) == "float64":
            sampled_values = [round(sample, 2) for sample in sampled_values]
        samples += f"{df_[column].name} ({df[column].dtype}): NaN-freq [{nan_freq}%], Samples {sampled_values}\n"
    return _get_prompt(df, nl_prompt, how_many, samples=samples, dag_summary=dag_summary)


def _dag_summary_info(gyyre_dag_summary):
    if gyyre_dag_summary is not None:
        return (
            f" for a {gyyre_dag_summary.task_type} task, predicting `{gyyre_dag_summary.target_name}` "
            f"with {gyyre_dag_summary.model}"
        )
    return ""


def _add_memorized_history(
    gyyre_memory: list[dict[str, Any]] | None,
    messages: list[dict[str, str]],
    generated_code: list[str],
) -> None:
    if gyyre_memory is not None and len(gyyre_memory) > 0:
        current_accuracy = 0.0
        current_roc = 0.0

        for memory_line in gyyre_memory:
            memorized_code = memory_line["update"]
            memorized_accuracy = memory_line["score"]
            # TODO also compute and provide ROC AUC
            memorized_roc = memory_line["score"]

            improvement_acc = memorized_accuracy - current_accuracy
            improvement_roc = memorized_roc - current_roc

            if improvement_roc + improvement_acc >= 0.0:
                generated_code.append(memorized_code)
                add_feature_sentence = "The code was executed and changes to ´df´ were kept."
                current_accuracy = memorized_accuracy
                current_roc = memorized_roc
            else:
                add_feature_sentence = (
                    f"The last code changes to ´df´ were discarded. "
                    f"(Improvement: {improvement_roc + improvement_acc})"
                )

            messages += [
                {"role": "assistant", "content": memorized_code},
                {
                    "role": "user",
                    "content": f"Performance after adding feature ROC {memorized_roc:.3f}, "
                    f"ACC {memorized_accuracy:.3f}. {add_feature_sentence}\nNext codeblock:\n",
                },
            ]


def _try_to_execute(df: pd.DataFrame, code_to_execute: str) -> None:
    df_sample = df.head(100).copy(deep=True)
    columns_before = df_sample.columns
    # print("-" * 120)
    # print(code_to_execute)
    # print("-" * 120)
    df_sample_processed = _safe_exec(code_to_execute, "df", safe_locals_to_add={"df": df_sample})
    columns_after = df_sample_processed.columns
    new_columns = sorted(set(columns_after) - set(columns_before))
    removed_columns = sorted(set(columns_before) - set(columns_after))
    print(
        f"\t> Computed {len(new_columns)} new feature columns: {new_columns}, "
        f"removed {len(removed_columns)} feature columns: {removed_columns}"
    )


# pylint: disable=too-many-ancestors
class LLMFeatureGenerator(
    BaseEstimator, TransformerMixin, GyyreContextAwareMixin, GyyrePrefittableMixin, GyyreOptimisableMixin
):
    def __init__(
        self,
        nl_prompt: str,
        how_many: int,
        gyyre_dag_summary: DagSummary | None | DataOp = None,
        gyyre_prefitted_state: dict[str, Any] | None | DataOp = None,
        gyyre_memory: list[dict[str, Any]] | None | DataOp = None,
    ) -> None:
        self.nl_prompt = nl_prompt
        self.how_many = how_many
        self.gyyre_dag_summary = gyyre_dag_summary
        self.gyyre_prefitted_state = gyyre_prefitted_state
        self.gyyre_memory = gyyre_memory

        self.generated_code_ = []

    def state_after_fit(self):
        return {"generated_code": self.generated_code_}

    def memory_update_from_latest_fit(self):
        if self.generated_code_ is not None and len(self.generated_code_) > 0:
            return self.generated_code_[-1]
        return ""

    def fit(self, df: pd.DataFrame, y=None, **fit_params):  # pylint: disable=unused-argument
        prompt_preview = self.nl_prompt[:40].replace("\n", " ").strip()

        if self.gyyre_prefitted_state is not None:
            print(f"--- Using provided state for gyyre.with_sem_features('{prompt_preview}...', {self.how_many})")
            self.generated_code_ = self.gyyre_prefitted_state["generated_code"]
            return self

        print(
            f"--- Fitting gyyre.with_sem_features('{prompt_preview}...', {self.how_many})"
            f"{_dag_summary_info(self.gyyre_dag_summary)}"
        )

        messages = []
        for attempt in range(1, _MAX_RETRIES + 1):
            code = ""

            try:
                prompt = _build_prompt_from_df(df, self.nl_prompt, self.how_many, self.gyyre_dag_summary)

                if attempt == 1:
                    messages += [{"role": "system", "content": _SYSTEM_PROMPT}, {"role": "user", "content": prompt}]
                    _add_memorized_history(self.gyyre_memory, messages, self.generated_code_)

                code = _generate_python_code_from_messages(messages)
                code_to_execute = "\n".join(self.generated_code_)
                code_to_execute += "\n\n" + code

                _try_to_execute(df, code_to_execute)

                self.generated_code_.append(code)
                break
            except Exception as e:  # pylint: disable=broad-except
                print(f"\t> An error occurred in attempt {attempt}:", e)
                messages += [
                    {"role": "assistant", "content": code},
                    {
                        "role": "user",
                        "content": f"Code execution failed with error: {type(e)} {e}.\n "
                        f"Code: ```python{code}```\n Generate next feature (fixing error?):\n```python\n",
                    },
                ]

        return self

    def transform(self, df):
        check_is_fitted(self, "generated_code_")
        code_to_execute = "\n".join(self.generated_code_)
        df = _safe_exec(code_to_execute, "df", safe_locals_to_add={"df": df})
        return df


class WithSemFeaturesCaafe(WithSemFeaturesOperator):
    def generate_features_estimator(self, data_op: DataOp, nl_prompt: str, name: str, how_many: int):
        gyyre_dag_summary = skrub.var(f"gyyre_dag_summary__{name}", None)
        gyyre_prefitted_state = skrub.var(f"gyyre_prefitted_state__{name}", None)
        gyyre_memory = skrub.var(f"gyyre_memory__{name}", [])

        return LLMFeatureGenerator(
            nl_prompt,
            how_many,
            gyyre_dag_summary=gyyre_dag_summary,
            gyyre_prefitted_state=gyyre_prefitted_state,
            gyyre_memory=gyyre_memory,
        )



from typing import Any

import numpy as np
import skrub
from skrub import DataOp

from gyyre.optimisers._dag_summary import _summarise_dag


def greedy_optimise_semantic_operator(
    dag_sink: DataOp,
    operator_name: str,
    num_iterations: int,
    scoring: str = "accuracy",
    cv: int = 3,
) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
    memory = []
    states = []

    print("--- COMPUTING DAG SUMMARY for context-aware optimisation ---")
    dag_summary = _summarise_dag(dag_sink)
    # print(dag_summary)

    for iteration in range(num_iterations):
        print(f"---ITERATION {iteration} -> Fitting with memory ---")
        learner = dag_sink.skb.make_learner(fitted=False)

        env = dag_sink.skb.get_data()
        env[f"gyyre_dag_summary__{operator_name}"] = dag_summary
        env[f"gyyre_memory__{operator_name}"] = memory

        learner.fit(env)

        print(f"---ITERATION {iteration} -> Evaluation ---")
        op_fitted = learner.find_fitted_estimator(operator_name).transformer_
        op_state = op_fitted.state_after_fit()
        states.append(op_state)

        env = dag_sink.skb.get_data()
        env[f"gyyre_prefitted_state__{operator_name}"] = op_state

        op_memory_update = op_fitted.memory_update_from_latest_fit()

        cv_results = skrub.cross_validate(learner, env, cv=cv, scoring=scoring)
        mean_score = np.mean(cv_results["test_score"])

        memory.append(
            {
                "update": op_memory_update,
                "score": float(mean_score),
            }
        )

        print(f"---ITERATION {iteration} -> {scoring}: {mean_score}")
    return memory, states
